#include "Tests.h"
#include "treefunctions.h"
#include <sstream>

using namespace std;
ENABLE_TESTS();

/* Build a test tree to use for remaining tests.
 * For more robust tests, we should create a variety
 * of different trees, but for this case we will just
 * test with this one.
 */
TreeNode *buildTestTree() {
    /*         9
     *       /   \
     *     6       14
     *    / \     / \
     *  -3   7  11   19
     */
    // build a tree by connecting nodes
    TreeNode *root = new TreeNode(9);
    root->left = new TreeNode(6);
    root->right = new TreeNode(14);
    root->left->left = new TreeNode(-3);
    root->left->right = new TreeNode(7);
    root->right->left = new TreeNode(11);
    root->right->right = new TreeNode(19);
    return root;
}

// Should just print the empty string
ADD_TEST("print empty tree") {
    TreeNode *empty = nullptr;
    ostringstream stream;
    print(empty, stream);
    expect(stream.str() == "");
}

// Should print the element followed by a newline
ADD_TEST("print one node tree") {
    TreeNode *oneNode = new TreeNode(42);
    ostringstream stream;
    print(oneNode, stream);
    expect(stream.str() == "42\n");
    deleteTree(oneNode);
}

// Should print each element on its own line, and a newline at the end
ADD_TEST("print test tree") {
    TreeNode *root = buildTestTree();
    ostringstream stream;
    print(root, stream);
    expect(stream.str() == "-3\n6\n7\n9\n11\n14\n19\n");
    deleteTree(root);
}

// Should be 0
ADD_TEST("test empty tree size") {
    expect(size(nullptr) == 0);
}

// Should be 1
ADD_TEST("test single node size") {
    TreeNode *root = new TreeNode(1);
    expect(size(root) == 1);
    deleteTree(root);
}

// Size should be total number of nodes in the tree
ADD_TEST("test tree size") {
    TreeNode *root = buildTestTree();
    expect(size(root) == 7);
    deleteTree(root);
}

// Contains on empty tree should always be false
ADD_TEST("contains on empty tree") {
    expect(!contains(nullptr, 2));
}

// Tree should always contain its root element
ADD_TEST("contains root") {
    TreeNode *root = new TreeNode(5);
    expect(contains(root, 5));
    deleteTree(root);
}

ADD_TEST("contains leaf element") {
    TreeNode *root = buildTestTree();
    expect(contains(root, -3));
    expect(contains(root, 7));
    expect(contains(root, 11));
    expect(contains(root, 19));
    deleteTree(root);
}

ADD_TEST("contains middle element") {
    TreeNode *root = buildTestTree();
    expect(contains(root, 6));
    expect(contains(root, 14));
    deleteTree(root);
}

// Should all be false
ADD_TEST("contains on missing elements") {
    TreeNode *root = buildTestTree();
    expect(!contains(root, 87));
    expect(!contains(root, 0));
    expect(!contains(root, 60));
    expect(!contains(root, -1));
    deleteTree(root);
}

// Height should be 0
ADD_TEST("height of empty tree") {
    expect(height(nullptr) == 0);
}

// Height should be 1
ADD_TEST("height of one node") {
    TreeNode *root = new TreeNode(1);
    expect(height(root) == 1);
    deleteTree(root);
}

// Height should be max number of nodes to reach bottom
ADD_TEST("height of multi-level tree") {
    TreeNode *root = buildTestTree();
    expect(height(root) == 3);
    deleteTree(root);
}

// Deleting nullptr should do nothing
ADD_TEST("deleting empty tree shouldn't do anything") {
    TreeNode::instances = 0;
    deleteTree(nullptr);
    expect(TreeNode::instances == 0);
}

// Make sure the node is deleted
ADD_TEST("deleting single node should just delete one node") {
    TreeNode::instances = 0;
    TreeNode *root = new TreeNode();
    expect(TreeNode::instances == 1);
    deleteTree(root);
    expect(TreeNode::instances == 0);
}

// Make sure every node allocated is deleted
ADD_TEST("tree deletion shouldn't leak any memory") {
    TreeNode::instances = 0;
    TreeNode *root = buildTestTree();
    expect(TreeNode::instances = 7);
    deleteTree(root);
    expect(TreeNode::instances == 0);
}
